static char *cvs_id =" $Id: spec2grace.c,v 1.19 2006/05/10 06:47:06 berar Exp $ ";
/* spec2grace
 *    to view specfile using grace
 *    can be used only to extract subfiles
 *
 * jf berar d2am
 *          1994-95            : view specfile using xmgr (3.0) with -npipe option as viewer
 *          modified 13Dec95   : col negativ and comments in middle of scan
 *          modified 23Jan95   : Normalized
 *          modified 31Jan95   : comments, write to file
 *          modified 30Jan96   : error bars
 *          modified 20Mar96   : add ZERO=1.0e-6 to counts to avoid 0 in log
 *          modified  2Mar97   : spec2xmgr.bat auto batch file
 *          modified  5Sep97   : ini error bars, W write columns C using format F
 *          modified 19Sep97   : reopen xmgr if possible
 *          modified 25Aug99   : spec2xmgr.bat uses GETP (4.12) and not READ BATCH(3.0)
 *          modified 17Jul00   : scan_header corrected for eof, default name for w, W
 *                             : user_index = index+1, scan_loop  
 *          modified 27Mar01   : compilation gcc linux ... xmgrace(grace5.1.1) ok
 *          xmgrace now becomes standard, if no GRACE library : define NOGRACE
 *          modified 30Mai01   : modified for grace, renamed spec2grace
 *                             no AUTOSCALE : does not work in grace5.1.3
 *          modified 23Jul01   : title, hard print...
 *          modified 12Dec02   : legends, autoscale and purge works...
 *          modified 20Feb03   : nearest scan, %g format to grace, autoXcolumn ...
 *          modified 21Feb03   : grace5.1.11 READ XY instead of READ BLOCK + UPDATEALL 
 *                              mygets (readline...) functions
 *          modified 27Feb03   : grace5.1.12, args as first commands, legends
 *          modified 27Jan06   : gnuplot support for mesh... (B Caillot)
 *          modified 24Feb06   : norm_ref value defined
 *          modified 09May06   : protect killing X server in atexit function
 *          modified 10May06   : check_index(inf==NULL) return 0, no more Segmentation fault
*/
/* ------------------------------------------------------ */
#define RELEASE "spec2grace, rel 2.11"
#define PRG_DOC "spec2grace.htlm"
/* ------------------------------------------------------ */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <stdarg.h>
#include <ctype.h>
#include <time.h>
#include <sys/wait.h>
/* ------------------------------------------------------ */
#include "mygets.h"
/* ------------------------------------------------------ */
#ifndef NOGRACE
/* the path is defined in Makefile */
#include <grace_np.h> 
#endif
/* ------------------------------------------------------ */
extern int errno;
/* ------------------------------------------------------ */
#define NDEBUG  1
/* ------------------------------------------------------ */
#define ZERO 1.0e-6
#define XCOL 1
#define DCOL -1
#define AUTOSCALE 1
#define GNUPLOT_CMD_FILE   "/usr/local/bin/spec2graceGnuplot.cmd"
/* ------------------------------------------------------ */
char *prgname;

int xcol = XCOL;
int xcol2 = XCOL;
int ycol = DCOL;
int zcol = DCOL-1;
int ncol = 0;
int xcol_used;
int x1col_used;
int x2col_used;

int mesh =0;

int autoscale = AUTOSCALE;
int normalize = 0;
int autopurge = 0;
int logscale  = 0;
int errorbar  = 0;
int comments  = 1;
int titleauto  = 1;
int legendauto  = 1;
int autoXcolumn = 1;

FILE * inf = NULL;
#define BUFSIZE 1024
char   buffer[BUFSIZE+1];
char   wformat[80];
char   wcolumns[80];   
char   ininame[80];
char  *inf_name = "NO FILE";

double ref_norm = 1.0;
/* ---------------------------------------------------------------------- */
typedef struct
{     int  scan;
      long hseek;
      long vseek;
      long cseek;
      long eseek;
      int  lpos;
      int  lines;
      int  ncol;
      int  Xcol;
      char *txt;
      }   scanRef;

#define INDEXALLOC 512      
scanRef * Index = NULL;
int    Indexed = 0;
int    IndexSize = 0;
long   IndexFileLength = 0;

FILE *my_stdout;
/* ---------------------------------------------------------------------- */
int index_open( char * name);
int index_dir( char * name);
int index_free();
int index_file();
int index_print();
int get_index();
int check_index();
char *spec_tag(int ref, char tag);

int xcol_select(int ref);
int x1col_select, x2col_select;
int scan_header(int ref);
int scan_print(int ref);
int scan_printf( int ref, char *name );
int scan_write( int ref, char *name );
int loop( int ref, char *name);
int mesh_edf( int ref, char *name );
int mesh_tiff( int ref, char *name );
static double *mesh_array(double *dmi,double *dma, scanRef *lIndex);
/* ------------------------------------------------------ */
#ifndef NOGRACE
#define GRACEPRINT "DEVICE \"PostScript\" PAGE SIZE 300,400"
char graceprint[256];
char graceWin[256];
char gracewinlegend[256];
int graceOrGnuplotSendScan      (int ref);
int gracesend_scan      (int ref);
int gnuplotsend_scan    (int ref);
int gracesend_print     (char *name);
int gracesend_batch     (char *name);
void gracewin_size      (void);
int gracesend_parameter (char *name);
void graceerror_function(const char *msg);
#endif
/* ------------------------------------------------------ */
int  getnum( char *c, int i);
int  getanum( char *c, int i);
int  get_ncol( char *buf, scanRef * lIndex);
double getdouble( char * buf, int pos, int* err);
double getadouble( char * buf, double def, int* err);
FILE *outputfile(int ref, int s, char *name);
/* ---------------------------------------------------------------------- */
int  put_cmds();
int  status();
char *prompt(int i);
int  interprete(char *c, int   i);
/* ---------------------------------------------------------------------- */
void usage( char * msg)
{
    if ( msg ) fprintf( stderr,"%s\n", msg);
    fprintf (stderr,"%s %s\n", RELEASE, __DATE__ );
    fprintf (stderr,"usage : %s [file [cmd [cmds]]] exex cmds on file then prompt\n", prgname );
    fprintf (stderr,"usage : %s -q file cmd [cmds]] quit after cmds\n", prgname );
    fprintf (stderr,"default value are X=%d Y=%d )\n", XCOL, DCOL);
    exit (3);
    }
/* ---------------------------------------------------------------------- */
int main ( int argc, char ** argv)
{
    FILE *infm;
    char *c;
    char   buffer[BUFSIZE+1];
    int  do_prompt=1;
    int i;
     
    my_stdout  = stdout;
    prgname = *argv;
    argc--; argv++;
#ifndef NOGRACE
    strcpy(graceprint,GRACEPRINT);
    strcpy(graceWin,"");
    strcpy(gracewinlegend,"");
#endif	 
    strcpy(wformat,"%8.4lf %8.4lf\n");
    strcpy(wcolumns,"1 -1");
    strcpy( ininame, prgname );
    strcat( ininame, ".ini" );
    infm = fopen ( ininame, "r" );
    if ( infm != NULL )
    {
    printf( "Using initialization file : %s\n", ininame);
    while ( !feof( infm ))
    {   int errval = 0;
        fgets( buffer, BUFSIZ, infm );
        if (strlen(buffer) < 1 ) continue;
        if ( feof( infm )) break;
        c = buffer;
        while ( *c == ' ' ) c++;
        switch ( *(c++) )
       {  case 'f' : index_open(c);
                     break;
          case 'x' : xcol = getanum(c , xcol);
                     printf (" using columns x=%d and y=%d for xmgr, z=%d as monitor\n", xcol, ycol, zcol);
                     break;
          case 'y' : ycol = getanum(c, ycol);
                     printf (" using columns x=%d and y=%d for xmgr, z=%d as monitor\n", xcol, ycol, zcol);
                     break;
          case 'z' : zcol = getanum(c, zcol);
                     printf (" using columns x=%d and y=%d for xmgr, z=%d as monitor\n", xcol, ycol, zcol);
                     break;
          case 'L' : logscale = getnum(c, logscale);
                     printf (" automatic log scale turned %s\n", logscale?"on":"off");
                     break;
          case 'N' : normalize = getnum(c, normalize);
                     printf (" automatic normalization turned %s\n", normalize?"on":"off");
                     break;
          case 'n' : errval=0;
                     ref_norm = getadouble(c, ref_norm, &errval);
                     printf (" normalization reference set to %lf\n", ref_norm);
                     break;
          case 'P' : autopurge = getnum(c, autopurge);
                     printf (" automatic purge turned %s\n", autopurge?"on":"off");
                     break;
          case '#' : comments = getnum(c, comments);
                     printf (" comments in file turned %s\n", comments?"on":"off");
                     break;
          case 'A' : autoscale = getnum(c, autoscale);
                     printf (" automatic autoscale turned %s\n", autoscale?"on":"off");
                     break;
          case 'X' : autoXcolumn = getnum(c, autoXcolumn);
                     printf (" automatic Xcolumn turned %s\n", autoXcolumn?"on":"off");
                     break;
          case 'E' : errorbar = getnum(c, errorbar);
                     printf (" automatic errorbar turned %s\n", errorbar?"on":"off");
                     break;
          case 'T' : titleauto = getnum(c, titleauto);
                     printf (" automatic title turned %s\n", titleauto?"on":"off");
                     break;
          case 'l' : legendauto = getnum(c, legendauto);
                     printf (" automatic legend turned %s\n", legendauto?"on":"off");
                     break;
          case 'F' : if (*c) strcpy( wformat, c);
                     wformat[strlen(wformat)-1]=0;
                     printf (" format used by the W command set to \"%s\"\n", wformat);
                     printf (" columns being \"%s\"\n", wcolumns);
                     break;
          case 'C' : if ( *c ) strcpy(wcolumns, c);
                     wcolumns[strlen(wcolumns)-1]=0;
                     printf (" columns used by the W command set to \"%s\"\n", wcolumns);
                     printf (" format being \"%s\"\n", wformat);
                     break;
#ifndef NOGRACE
          case 'H' : if ( *c ) strcpy(graceprint, c);
                     graceprint[strlen(graceprint)-1]=0;
                     printf (" parameters used by the h command set to \"%s\"\n", graceprint);
                     break;
          case 'U' : if ( *c ) strcpy(graceWin, c);
                     graceWin[strlen(graceWin)-1]=0;
                     printf (" parameters used by the W command set to \"%s\"\n", graceWin);
                     break;
          case 'u' : if ( *c ) strcpy(gracewinlegend, c);
                     gracewinlegend[strlen(gracewinlegend)-1]=0;
                     printf (" parameters used by the u command set to \"%s\"\n", gracewinlegend);
                     break;
#endif
              }
          }
    fclose (infm );
    }
    
   if (argc)
   {  if (strcmp(*argv,"-q")==0)
      {  do_prompt=0;
         argc--; argv++;
         if (argc<2) usage("cmds are required for -q");
         }
      index_open( *argv );
      argc--; argv++;
      }    

   if ( inf  )
   {  index_file();
      index_print();
      }         
   i = Indexed;
   while (argc)
   {   char *c;
       c = *argv;
       if ( c && *c )  
       {  printf("%s ARG='%s'\n", prompt(i),c);
          i=interprete(c,i);
          }
       argc--; argv++;
       if (i==-1) { do_prompt=argc=0; }
       }
   if (do_prompt)
   {   char *c, *line_read = (char *)NULL;

       status();   
       do   
       {   if (line_read)  free (line_read);
           c = line_read = readline(prompt(i));
           if ( c && *c )  i=interprete(c,i);
           } while ( i >= 0 ); 
       }
   infm = fopen ( ininame, "w" );
   if ( infm != NULL ) 
   {   fprintf (infm, "f %s\n", inf_name );
       fprintf (infm, "x %d\ny %d\nz %d\n", xcol, ycol, zcol);
       fprintf (infm, "A %d\nP %d\nX %d\nL %d\nN %d\nE %d\nT %d\nl %d\n# %d\n", autoscale, autopurge, 
            autoXcolumn, logscale, normalize, errorbar, titleauto, legendauto, comments);
       fprintf (infm, "F%s\nC%s\n", wformat, wcolumns);
#ifndef NOGRACE
       fprintf (infm, "H%s\n", graceprint);
       fprintf (infm, "u%s\n", gracewinlegend);
       fprintf (infm, "U%s\n", graceWin);
#endif
       fclose  (infm );
       }
   }
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
int put_cmds( )
{
    printf("\n %s %s see also %s\nrecognized commands are\n", RELEASE, __DATE__, PRG_DOC );
#ifndef NOGRACE
    printf("  v : view indexed scan using xmgrace (or gnuplot for a mesh) (current, index or relative)\n");
    printf("  h : hard copy from xmgrace\n");
#endif
    printf("  S, c : scan header, comments  p : print current scan \n");
    printf("  t : print the index table     w : write current scan to a file (-=stdout)\n");
    printf("  i : set current index         s : set current scan\n");
    printf("  r : reindex the file          q : quit\n");
    printf("  f : use another file          d : set searched dir\n");
    printf("  ( : (start|istart,end,v|w|W|p|h|c|e)[fileloop]  loop  (3,5,w) (s4,7,v) ....\n");
    printf("  x, y , z : set xmgr columns x or y, or the monitor (z) (1..N or -1=last column\n\n");
    printf("  Rr : spec tag line r          X : toggle automatic column selection\n");
    printf("  e,m : mesh_export (edf,tiff)  \n");
#ifndef NOGRACE
    printf("  b to send a batch file        B to send a parameter file to xmgr\n"); 
    printf("    files 'spec2grace.bat' or 'spec2grace.par' are send automaticaly with the view command\n");
    printf("  A : toggle automatic AUTOSCALE in xmgrace\n");
    printf("  E : toggle error bar mode in data\n");
    printf("  L : toggle automatic LOG scale\n");
    printf("  N : toggle automatic NORMALIZATION,   n : set monitor reference value = %lf\n", ref_norm);
    printf("  P : toggle automatic PURGE in xmgrace\n");
    printf("  T : toggle automatic TITLE in xmgrace\n");
    printf("  l : toggle automatic LEGEND in xmgrace\n");
    printf("  H : hard copy parameters\n");
    printf("  u : set args for legeng, none=scannumber,T=content of #T, +T=scannumber+content of#T\n");
    printf("  U : set grace Window size\n");
#endif
    printf("  # : toggle writing comments in file\n");
    printf("  X : toggle automatic column selection\n");
    printf("  W : write to file according to the following specifications\n");
    printf("  F : format string, pe : \"%s\"\n", wformat);
    printf("  C : column numbers separate by spacespe, pe : \"%s\"\n", wcolumns); 
}
/* ---------------------------------------------------------------------- */
int status( )
{
    printf("\n %s, type ? to get list of commands\n", prgname );
    printf("\n current file is : %s with %d entries\n", inf_name, Indexed);
    printf(" setting is : x=%d, y=%d autoscale=%s autopurge=%s Xcolumn=%s\n errorbar=%s logscale=%s normalization=%s(z=%d,ref=%lf)\n title=%s  comments=%s\n", 
         xcol, ycol, autoscale?"on":"off", autopurge?"on":"off", autoXcolumn?"on":"off", 
         errorbar?"on":"off", logscale?"on":"off", 
         normalize?"on":"off",zcol, ref_norm, titleauto?"on":"off", comments?"on":"off");
    printf (" current W command parameters are \"C %s\", \"F %s\"\n", wcolumns, wformat);
#ifndef NOGRACE
    printf (" current H command parameters \"%s\" used for grace hardcopy\n", graceprint);
    printf (" current u command parameters \"%s\" used for grace legend\n", gracewinlegend);
    printf (" current U command parameters \"%s\" used for grace size\n", graceWin);
#endif
}
/* ---------------------------------------------------------------------- */
char *prompt(int i)
{  static char prompt_buffer[200];
   if ( Indexed == 0 )    sprintf(prompt_buffer,"%s >", inf_name);
   else sprintf(prompt_buffer,"%s (i=%d, s=%d) >", inf_name, i , Index[i-1].scan);
   return prompt_buffer;
   }
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
int interprete(char *c, int   i)
{  int   s;
   int errval = 0;
   switch ( *(c++) )
   {  case 'S' : i = getnum(c, i);
                 if ( check_index(i-1) != 0 ) scan_header(i-1);
                 break;
      case 'c' : i = getnum(c, i);
                 if ( check_index(i-1) != 0 ) scan_comment(i-1);
                 break;
      case 'w' : scan_write(i-1, c);
                 break;
      case 'p' : i = getnum(c, i);
                 if ( check_index(i-1) != 0 ) scan_print(i-1);
                 break;
      case 'R' : fprintf(my_stdout,"%s\n",spec_tag(i,*c));
                 break;
#ifndef NOGRACE
      case 'v' : i = getnum(c, i);
                 if ( check_index(i-1) != 0 ) graceOrGnuplotSendScan(i-1);
                 break;
      case 'b' : gracesend_batch(c);
                 break;
      case 'B' : gracesend_parameter(c);
                 break;
      case 'h' : gracesend_print(c);
                 break;
      case 'H' : if ( *c ) strcpy(graceprint, c);
                 printf ("parameters of the h (grace hardcopy) command set to \"%s\" \n", graceprint);
                 break;
      case 'U' : if ( *c ) strcpy(graceWin, c);
                 graceWin[strlen(graceprint)-1]=0;
                 printf (" parameters used by the U command set to \"%s\"\n", graceWin);
                 gracewin_size();
                 break;
      case 'u' : if ( *c ) strcpy(gracewinlegend, c);
                 gracewinlegend[strlen(graceprint)-1]=0;
                 printf (" parameters used by the u command set to \"%s\"\n", gracewinlegend);
                 break;
#endif
      case 'm' : i = getnum(c, i);
                 if ( check_index(i-1) != 0 ) mesh_tiff(i-1, c);
                 break;
      case 'e' : i = getnum(c, i);
                 if ( check_index(i-1) != 0 ) mesh_edf(i-1, c);
                 break;
      case 't' : index_print();
                 break;
      case 's' : s = Index[i-1].scan;
                 i = get_index(i, getnum(c, s)) + 1;
                 break;
      case 'i' : i = getnum(c, i);
                 if ( check_index(i-1) == 0 ) i = Indexed ;
                 break;
      case 'r' : index_free();
                 index_file();
                 i = Indexed;
                 break;
      case 'f' : if( index_open(c) ) index_file();
                 i = Indexed;
                 break;
      case 'd' : if( index_dir(c) );
                 break;
      case '(' :  i = loop(i, c);
	              break;           
      case 'q' : return -1;           
      case 'x' : xcol = getanum(c , xcol);
                 printf (" using columns x=%d and y=%d for xmgr, z=%d as monitor\n", xcol, ycol, zcol);
//                 c++;
//                 xcol2 = getanum(c , xcol+1);
//                 printf (" using columns x=%d and y=%d for xmgr, z=%d as monitor\n", xcol2, ycol, zcol);
                 break;
      case 'y' : ycol = getanum(c, ycol);
                 printf (" using columns x=%d and y=%d for xmgr, z=%d as monitor\n", xcol, ycol, zcol);
                 break;
      case 'z' : zcol = getanum(c, zcol);
                 printf (" using columns x=%d and y=%d for xmgr, z=%d as monitor\n", xcol, ycol, zcol);
                 break;
      case 'X' : autoXcolumn = !autoXcolumn;
                 printf (" automatic Xcolumn turned %s\n", autoXcolumn?"on":"off");
                 break;
#ifndef NOGRACE
      case 'A' : autoscale = !autoscale;
                 printf (" automatic autoscale turned %s\n", autoscale?"on":"off");
		           break;
      case 'T' : titleauto = !titleauto;
                 printf (" automatic title turned %s\n", titleauto?"on":"off");
                 break;
      case 'l' : legendauto = !legendauto;
                 printf (" automatic legend turned %s\n", legendauto?"on":"off");
                 break;
      case 'L' : logscale = !logscale;
                 printf (" automatic log scale turned %s\n", logscale?"on":"off");
                 break;
      case 'P' : autopurge = !autopurge;
                 printf (" automatic purge turned %s\n", autopurge?"on":"off");
                 break;
#endif
      case 'E' : errorbar = !errorbar;
                 printf (" error bars turned %s\n", errorbar?"on":"off");
                 break;
      case 'N' : normalize = !normalize;
                 printf (" automatic normalization turned %s\n", normalize?"on":"off");
                 break;
      case 'n' : ref_norm = getadouble(c, ref_norm, &errval);
                 printf (" normalization reference set to %lf\n", ref_norm);
                 break;
      case '#' : comments = !comments;
                 printf (" comments in file turned %s\n", comments?"on":"off");
                 break;
      case 'W' : scan_printf(i-1, c);
                 break;
      case 'F' : if ( *c ) strcpy(wformat, c);
                 printf (" format used by the W command set to \"%s\" with columns \"%s\"\n", wformat, wcolumns);
                 break;
      case 'C' : if ( *c ) strcpy(wcolumns, c);
		           printf (" columns used by the W command set to \"%s\" with format \"%s\"\n", wcolumns, wformat);
                 break;
      case '?' : put_cmds(); status(); break;
      default  : printf ("??? '%s' ???\n", c-1); status();
      }
    return i; 
    }
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
int getnum( char *c, int i)
{   int num = 1;
    int rel = 0;
    while ( *c )  switch (*c)
	 {  case ' '  : c++; break;
	    case '+'  : rel=1; c++; break;
	    case '-'  : rel=-1; ; c++; break;
       case '\n' : 
       case '\r' : 
       case 0    : return i + rel;
		 default   : errno = 0; num = atoi(c);
                   if ( errno ) return i;
                   if ( rel ) return i + rel*num;
						 return num;
		 }
	 return i + rel*num;
    }
/* ---------------------------------------------------------------------- */
int getanum( char *c, int i)
{
    int num;
    while ( *c == ' ' ) c++;
    if ( *c == 0 ) return i;
    errno = 0;
    num = atoi( c );
    if( num == 0) return i;
    if ( errno ) return i;
    return num;
    }
/* ---------------------------------------------------------------------- */
int get_ncol( char *buf, scanRef * lIndex)
{
   char format[100];
   double x;
   int i, pos;
   if (lIndex->ncol > 0) return lIndex->ncol; 
   pos = -1;
   do
   {   pos++;
       format[0] = 0;
       for ( i = 1; i < pos; i++ ) strcat( format, "%*f" );
       strcat( format, "%lf" );
       }
       while ( sscanf( buf, format, &x) == 1 );
   lIndex->ncol = pos-1;
   return  lIndex->ncol;
   }
/* ---------------------------------------------------------------------- */
double getdouble( char * buf, int pos, int* err)
{
   char format[100];
   double x;
   int i;
   format[0] = 0;
   if ( pos < 0 ) pos = 1+ncol+pos;
   for ( i = 1; i < pos; i++ ) strcat( format, "%*f" );
   strcat( format, "%lf" );
   *err = ( sscanf( buf, format, &x) == 1 ) ? 0 : -1 ;
   return x;
   }
/* ---------------------------------------------------------------------- */
double getadouble( char * buf, double defval, int* err)
{
   char format[100];
   double x=defval;
   *err = ( sscanf( buf, "%lf", &x) == 1 ) ? 0 : -1 ;
   return x;
   }
/* ---------------------------------------------------------------------- */
FILE *outputfile(int ref, int s, char *name)
{  char *c = name;
   FILE *tmp;
   char file_name[50];
	printf( " writing data on '%s'\n", c); 
	while ( *c == ' ') c++;
   if ( *c == '-' ) 
	{   tmp = my_stdout; 
		 printf( " writing data for (i=%d, s=%d) on 'stdout'\n", ref+1, s); 
	    } 
	else
	{  if ( strlen (c) )
	   {   char *d = c + ( strlen (c) -1);
          while (*d == ' ' ||  *d == '\n' ||  *d == '\r') { *d=0; d--; }
			 }
		if ( *c == '*' || *c == 0 ) 
	   {   sprintf(file_name,"scan%03d.%03d", s, ref+1);
	       c = file_name;
			 }
	    tmp = fopen( c, "w" );
		 printf( " writing data for (i=%d, s=%d) on '%s'\n", ref+1, s, c); 
	    } 
   return tmp;
	}		 
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/*     loop procedure                                                     */
/* ---------------------------------------------------------------------- */
int loop( int ref, char *name )
{    
   char buffer[BUFSIZE+1];
	FILE *redirect, *oldout;
   char *c, e;
	int i_end, n;
	int  i, s, index_mode=0;
	c = name;
	while ( *c==' ') c++;
	i = ref;
	s = Index[ref-1].scan;
	if ( ! isdigit(*c) ) switch (*c)
	{   case 'i' : index_mode = 1; 
	               sscanf(++c,"%d%n",&i,&n);
						c += n;
						if ( check_index(i-1) == 0 ) i = Indexed ;
						break;
	    case 's' : index_mode = 0; 
		            s = Index[ref-1].scan;
	               sscanf(++c,"%d%n",&s,&n);
						c += n;
                  i = get_index(ref, s) + 1;
                  break;
	    case ',' : index_mode = 0; break;
		 default : fprintf(stderr," Unable to interpret loop '%s' not starting with scan or index!\n",
		           name);
					  return 0;
		 } else
	{  index_mode = 0; 
		s = Index[ref-1].scan;
	   sscanf(c,"%d%n",&s,&n);
		c += n;
      i = get_index(ref, s) + 1;
      }
	sscanf(c,",%d%n", &i_end, &n);
	c += n;
	while ( *c == ' ' ) c++;
	if ( *c == ',') c++;
	while ( *c == ' ') c++;
	s = Index[i-1].scan;
	e = *c;
	c++;
	buffer[0]=0;
	sscanf(c,")%s", buffer);
	if (buffer[0] )
	{    printf(" Redirecting stdout to %s\n", buffer);
	     redirect = fopen(buffer,"w");
		  my_stdout = redirect;
		  } else redirect = NULL;
	printf(" loop '(%s' from (i=%d, s=%d) to '%s' %d command=%c\n", name,
	    i, s, index_mode==1?"index":"scan", i_end, e);  
   do
	{   if ( check_index(i-1) == 0 ) break;
	    switch (e)
	    {   case 'w' : scan_write(i-1,""); break;
	        case 'W' : scan_printf(i-1,""); break; 
#ifndef NOGRACE
	        case 'v' : graceOrGnuplotSendScan(i-1); break;
#endif
	        case 'p' : scan_print(i-1); break;
	        case 'h' : scan_header(i-1); break;
	        case 'c' : scan_comment(i-1); break;
	        case 'e' : scan_header(i-1);  scan_print(i-1); break;
		     default  : fprintf(stderr," Unable to interpret loop '%s' invalid command '%s'!\n", 
			                name, c );  
					      break;	
           }
		  if ( (index_mode == 1 && i >= i_end) || (index_mode == 0 && s == i_end)) break; 
		  i++;	  
	     s = Index[i-1].scan;
		  } while(1);
	if (redirect) 
	{   fclose(redirect);
	    my_stdout = stdout;
		 }
	if (i > Indexed) i = Indexed;
	return i;	
   }
/* ---------------------------------------------------------------------- */
/*     specfile indexing procedures                                       */
/* ---------------------------------------------------------------------- */
int index_open( char *name )
{
   char buffer[BUFSIZE+1];
   char *e;
   char *f = name;
   while ( *f == ' ') f++;
   e = f;
   if ( *f == '~' ) 
   {    strcpy(buffer, getenv("HOME"));
        f++;
        strcat(buffer, f);
        }
   else strcpy(buffer, f);
   f = strchr(buffer, '\n' );
   if ( f ) *f = 0;
        
   inf = fopen( buffer, "r");
   if ( inf == NULL ) 
   {    printf( " file %s not found\n", buffer );
        inf_name = "NO FILE";
        Indexed = 0;
        return 0;
        }
   f = strchr(e, '\n' );
   if ( f ) *f = 0;
   f = malloc( strlen(e) + 1);
   strcpy( f, e);
   inf_name = f;
   return 1;    
   }     
/* ---------------------------------------------------------------------- */
int index_dir( char *name )
{  char buffer[BUFSIZE+1];
   char *f = name;
   while ( *f == ' ') f++;
   if ( *f == '~' ) 
   {    strcpy(buffer, getenv("HOME"));
        f++;
        if (*f) strcat(buffer, f);
        }
   else strcpy(buffer,f);
   chdir(buffer);     
   printf( "current working dir : %s\n", getcwd(buffer, BUFSIZE)  );
   return 0;
   }     
/* ---------------------------------------------------------------------- */
int index_free()
{   int i;
    scanRef * lIndex;
    for ( i = 0, lIndex = Index; i < Indexed; i++, lIndex++) 
       free (lIndex->txt);
    free( Index); 
    IndexSize = 0;
    Index = NULL;
    }
/* ---------------------------------------------------------------------- */
int index_file()
{  scanRef * lIndex;
   int lpos = 0;
   int lines;
   int i, l;
   int sz;
   long pseek;
   long lseek;
   enum { waithead, waitdata, waitend } state;
   char *b = buffer;
   char *c;
   double x;
   
   if ( Index == NULL ) Index = malloc(INDEXALLOC * sizeof(scanRef));
   if ( Index != NULL ) IndexSize = INDEXALLOC;
   else 
   {   printf("unable to allocate enough memory, exit!\n");
       exit(3);
       }
   lIndex = Index;
   Indexed = 0;
   state = waithead;
   
   lseek = 0;
   fseek( inf, 0, SEEK_SET );
   while ( !feof( inf) )
   {    pseek = ftell( inf);
        fgets( buffer, BUFSIZE, inf );
        lpos++;
        
        if ( *b == '#' )
        {   if ( state == waitend && buffer[1] == 'S' )
            {   /* lIndex->lines = lpos - l;*/
                lIndex++;
                state = waithead;
                }
            if ( state == waitdata )
            {  
#ifdef DEBUG
    printf("# waitdata %s", buffer);
#endif
                if (buffer[1] == 'S' )
                {   /* lIndex->lines = 0; */
                    lIndex++;
                    state = waithead;
                    }
                if (buffer[1] == 'N' )
                {   lIndex->ncol = 0;
                    sscanf(buffer+2,"%d",&lIndex->ncol);
                    }
                }
            if ( state == waithead )
            {   
#ifdef DEBUG    
printf("# waithead %s", buffer);
#endif
                if (buffer[1] == 'S' )
                {        if (Indexed) Index[Indexed-1].eseek = pseek;
                         if ( Indexed == IndexSize )
                    {    Index = realloc(Index, (IndexSize+INDEXALLOC) * sizeof(scanRef));
                         lIndex = Index + IndexSize;
                         if ( Index != NULL ) IndexSize += INDEXALLOC;
                         else 
                         {   printf("unable to allocate enough memory, exit!\n");
                             exit(3);
                             }
                         }
                    lIndex->scan = atoi(buffer+2);
                    lIndex->hseek = pseek;
                    lIndex->vseek = 0;
                    lIndex->lines = 0;
                    lIndex->ncol = 0;
                    lIndex->lpos = lpos;
                    lIndex->cseek = lseek;
                    lines = 0;
                    c = buffer+2;
                    while ( *c == ' ' ) ++c;
                    while ( isdigit(*c) ) ++c;
                    while ( *c == ' ' ) ++c;
                    sz = strlen(c);
                    lIndex->txt = malloc( 1 + sz);
                    strcpy( lIndex->txt, c);
                    lIndex->txt[sz-1] = '\n';
                    Indexed++;
                    state = waitdata;
                    }
                }
            }
        else if ( *b  == 0 )
        {   
#ifdef DEBUG    
printf(" *b  == 0 %s", buffer);
#endif
            if ( state == waitend )
            {   /* lIndex->lines = lpos - l;*/
                lIndex++;
                lseek = pseek;
                state = waithead;
                }
            }
        else 
        {   if ( state == waitdata )
            {   
#ifdef DEBUG    
printf(" else waitdata  %s", buffer);
#endif
                l = lpos;
                lseek = lIndex->vseek = pseek;
                state = waitend;
                }
            if ( state == waitend && sscanf(buffer,"%lf",&x)==1 )
            {   
#ifdef DEBUG    
printf(" else waitend %d : %s", lIndex->lines, buffer);
#endif
                lIndex->lines++;
                }
#ifdef DEBUG    
else printf(" else  %s", buffer);
#endif
            }
        }
/*    if ( state == waitend )  lIndex->lines = lpos - l; */
    fseek( inf, 0, SEEK_END );
    lIndex->eseek = IndexFileLength = ftell( inf );
/*   for ( i = 0, lIndex = Index; i < Indexed; i++, lIndex++)  
               if ( lIndex->lines ) lIndex->lines-- ;*/
    printf("\n current file is : %s with %d entries\n", inf_name, Indexed);
    return Indexed;
    }        
/* ---------------------------------------------------------------------- */
int index_print()
{    
   scanRef * lIndex;
   int     ui, i;
   for ( i = 0, ui=1, lIndex = Index; i < Indexed; ui++, i++, lIndex++)  
   {   printf(" %4d  Scan %4d  Lines %4d N %3d (%d)", ui, 
          lIndex->scan, lIndex->lines, lIndex->ncol, lIndex->lpos );
       if  (lIndex->lines && lIndex->txt )   printf(" %s", lIndex->txt);
     else  printf("\n");
       }   
   }
/* ---------------------------------------------------------------------- */
int get_index( int i_previous, int s)
{    
   scanRef * lIndex;
   int     s_previous=Index[i_previous-1].scan;
   int     i;
   int     old;
   if ( Indexed == 0 ) return -1;
   if ( s==s_previous) return i_previous;
   if ( s>s_previous)
   {  for ( i = i_previous, lIndex = Index+i_previous; i < Indexed; i++, lIndex++)
        if (lIndex->scan == s ) return i;  
      for ( i = 0, lIndex = Index; i < i_previous; i++, lIndex++)
        if (lIndex->scan == s ) return i; 
      }
   else
   {  for ( i = i_previous-1, lIndex = Index+i_previous-1; i >= 0; i--, lIndex--)
        if (lIndex->scan == s ) return i;  
      for ( i = Indexed-1, lIndex = Index+Indexed; i >= i_previous; i--, lIndex--)
        if (lIndex->scan == s ) return i; 
      }
   fseek( inf, 0, SEEK_END );
   if ( IndexFileLength != ftell( inf ))
   {    printf(" scan %d not found in the index_table, reindexing\n");
        old = Indexed;
        index_free();
        index_file();
        for ( i = old, lIndex = Index+old; i < Indexed; i++, lIndex++)
        if (lIndex->scan == s ) return i;  
        };
   printf(" scan %d not found in the index_table, ???\n", s);
   return i_previous;      
   }
/* ---------------------------------------------------------------------- */
int check_index ( int i)
{    if (inf == NULL)
     {   printf("NO FILE (check_index)\n");
         return 0;
         }
     if ( i >= Indexed - 1 )
     {    fseek( inf, 0, SEEK_END );
          if ( IndexFileLength != ftell( inf ))
          {    printf("reindexing\n");
               index_free();
               index_file();
               }
           }     
     if ( i < Indexed ) return 1;
     printf(" %d is not a valid index\n", i);
     return 0;
     }
/* ---------------------------------------------------------------------- */
/*     specfile data procedures                                           */
/* ---------------------------------------------------------------------- */
static int in_mesh=0;
static int valid_xcol(int x, int r, int s, char *t)
{   if (in_mesh)  printf (" mesh %d ",in_mesh);
    printf ("using xcol=%d column selected for (%d,%d) %s", 
       x, r, s, t); 
    return x;
    }
static int unvalid_xcol(int r, int s, char *t)
{   in_mesh=0;
    printf ("using xcol=%d : unable to select column for (%d,%d) %s\n", 
       xcol, r, s, t); 
   return xcol;    
   }
int xcol_select( int ref )
{  char cmd[20], mot[10], mot2[10];
   double hstart, hstop, kstart, kstop, lstart, lstop;
   int  hstep, kstep;
   scanRef * lIndex;
   if (!autoXcolumn) return xcol;
   in_mesh=0;
   lIndex = Index + ref;
   if ( sscanf(lIndex->txt," %s %lf %lf %lf %lf %lf %lf", &cmd, &hstart, &hstop, &kstart, &kstop, &lstart, &lstop) == 7)
   {   double dh,dk,dl;
       if (strcmp(cmd,"hklscan")!=0) return unvalid_xcol(ref,lIndex->scan, lIndex->txt);
       dh = fabs(hstart-hstop);
       dk = fabs(kstart-kstop);
       dl = fabs(lstart-lstop);
       if ( dh >= dk && dh >= dl) return valid_xcol(1,ref,lIndex->scan, lIndex->txt);
       if ( dk >= dh && dk >= dl) return valid_xcol(2,ref,lIndex->scan, lIndex->txt);
       if ( dl >= dh && dl >= dk) return valid_xcol(3,ref,lIndex->scan, lIndex->txt);
       }
   if  ( sscanf(lIndex->txt," %s %s %lf %lf %lf %lf", &cmd, &mot, &hstart, &hstop, &kstart, &kstop) == 6)   
   {   /* if (strcmp(cmd,"ascan")!=0 &&strcmp(cmd,"a2scan") ) return unvalid_xcol(ref,lIndex->scan, lIndex->txt);*/
       return valid_xcol(1, ref,lIndex->scan, lIndex->txt);
       }
   if ( sscanf(lIndex->txt," %s %s %lf %lf %d %s %lf %lf %d", &cmd, &mot, &hstart, &hstop, &hstep, 
              &mot2, &kstart, &kstop, &kstep) == 9)
   {   double dh,dk;
       if (strcmp(cmd,"hklmesh")!=0) return unvalid_xcol(ref,lIndex->scan, lIndex->txt);
       dh = fabs(hstart-hstop);
       dk = fabs(kstart-kstop);
       in_mesh=hstep;
       if ( hstep >= kstep && dh > 0) 
       {   if (strcmp(mot,"H")==0) return valid_xcol(1,ref,lIndex->scan, lIndex->txt);
           if (strcmp(mot,"K")==0) return valid_xcol(2,ref,lIndex->scan, lIndex->txt);
           if (strcmp(mot,"L")==0) return valid_xcol(3,ref,lIndex->scan, lIndex->txt);
           return unvalid_xcol(ref,lIndex->scan, lIndex->txt);
           }
       if ( kstep >= hstep && dk >0) 
       {   if (strcmp(mot2,"H")==0) return valid_xcol(1,ref,lIndex->scan, lIndex->txt);
           if (strcmp(mot2,"K")==0) return valid_xcol(2,ref,lIndex->scan, lIndex->txt);
           if (strcmp(mot2,"L")==0) return valid_xcol(3,ref,lIndex->scan, lIndex->txt);
           return unvalid_xcol(ref,lIndex->scan, lIndex->txt);
           }
       }
   return unvalid_xcol(ref,lIndex->scan, lIndex->txt);
   }
/* ---------------------------------------------------------------------- */
int scan_header( int ref )
{
   scanRef * lIndex;
   int npts = -1;
   int pos = 0;
   
   lIndex = Index + ref;
   fseek( inf, lIndex->hseek, SEEK_SET );
   do
   {    buffer[0]=0;
	     if (feof( inf ) ) break;
		  fgets( buffer, BUFSIZE, inf);
        if ( buffer[0] == '#' ) fprintf(my_stdout, "%s", buffer);
        } while ( buffer[0] == '#' );
    return ref;    
    }
/* ---------------------------------------------------------------------- */
int scan_comment( int ref )
{
   scanRef * lIndex;
   char *c;
   int npts = -1;
   int pos = 0;
   
   lIndex = Index + ref;
   fseek( inf, lIndex->cseek, SEEK_SET );
   do
   {    fgets( buffer, BUFSIZE, inf);
        if ( isdigit(buffer[0]) || buffer[0] == '-' || buffer[0] == '+' ) continue;
         fprintf(my_stdout, "%s", buffer);
        } while ( ftell(inf) < lIndex->eseek );
    return ref;    
    }
/* ---------------------------------------------------------------------- */
char *spec_tag(int ref, char tag)
{
   scanRef * lIndex;
   char *c;
   int npts = -1;
   int pos = 0;
   
   lIndex = Index + ref;
   fseek( inf, lIndex->cseek, SEEK_SET );
   do
   {    fgets( buffer, BUFSIZE, inf);
        if ( buffer[0]!= '#') continue;
        if ( buffer[1]  == 'S' ) 
        {  sscanf(buffer+2,"%d", &pos);
           if (pos>lIndex->scan) return "";
           }
        if ( buffer[1]  == tag ) 
        {  c=strchr(buffer,'\n');
           *c=0;
           return buffer;
           }
        } while ( ftell(inf) < lIndex->eseek );
    buffer[0]=0;
    return buffer;    
    }
/* ---------------------------------------------------------------------- */
int scan_print( int ref )
{  scanRef * lIndex;
   int i;
   lIndex = Index + ref;
   fseek( inf, lIndex->vseek, SEEK_SET );
   if ( lIndex->lines < 1 )
   {     fprintf(my_stdout, " No data to plot lines = %d\n", lIndex->lines );
         return ref;
         }
   for ( i = 0; i < lIndex->lines; i++)
   {    fgets( buffer, BUFSIZE, inf);
        fprintf(my_stdout, "%s", buffer);
        if ( buffer[0] == '#' ) continue;
        if (buffer[1] == 'S' || buffer[1] == 'R') break; 
        };
    return ref;    
    }

/* ---------------------------------------------------------------------- */
int scan_write( int ref, char *name )
{
   scanRef * lIndex;
   int i, s;
   int errval;
   FILE *tmp;
   double x, y, z, r, e;
       
   lIndex = Index + ref;
   if ( lIndex->lines < 1 )
   {     printf(" No data to write = %d\n", lIndex->lines );
         return ref;
         }
   s = Index[ref].scan;
   tmp = outputfile(ref,s,name);
   fseek( inf, lIndex->cseek, SEEK_SET );
   xcol_used = xcol_select(ref);
   do
   {    fgets( buffer, BUFSIZE, inf);
        if ( isdigit(buffer[0]) || buffer[0] == '-' || buffer[0] == '+' ) continue;
        if ( buffer[0] == '#' ) if (comments) fprintf( tmp,  "%s", buffer);
        } while ( ftell(inf) < lIndex->eseek );
   if (comments) fprintf( tmp, "#\n#spec2grace extracted columns %d, %d %d\n", xcol_used, ycol, zcol);
   if (comments) fprintf( tmp, "#\n#  x    y   %s n  %s %s\n#\n", errorbar?"s(y)":"",
          normalize?"y/n":"", errorbar&&normalize?"s(y/n)":"");

   fseek( inf, lIndex->vseek, SEEK_SET );
   for ( i = 0; i < lIndex->lines; i++)
   {     fgets( buffer, BUFSIZE, inf);
         if ( buffer[0] == '#' ) 
         {  i--;
            if (comments) fprintf(tmp, "%s", buffer);
            if (buffer[1] == 'S' || buffer[1] == 'R') break; 
            continue;
            }
         if ( i==0 ) ncol = get_ncol( buffer, lIndex);
         x = getdouble( buffer, xcol_used, &errval );
         if ( errval ) break;
         y = getdouble( buffer, ycol, &errval );
         z = getdouble( buffer, zcol, &errval );
         fprintf(tmp, "%15.5lf %15.7lg", x, y);
         if ( errorbar ) 
         {   if ( y > 1) e = sqrt(y); else e = 1;
             fprintf(tmp, " %15.7lg", e);
             }
         fprintf(tmp, " %15.7lg", z);
         if ( normalize ) 
         {   r = 0;
             if ( z >= 1 ) 
             {    r =  ref_norm * y / z;
                  fprintf(tmp, " %15.7lg", r);
                  if (errorbar) 
                  {    e = ref_norm *sqrt( e*e + ref_norm *y*y/z)/z;
                       fprintf(tmp, " %15.7lg", e);
                   }
             }
         }
         fprintf(tmp, "\n");
         };
    if (tmp!=stdout) fclose(tmp);     
    return ref;    
    }

/* ---------------------------------------------------------------------- */
int scan_printf( int ref, char *name )
{
   scanRef * lIndex;
   int i, j, s ;
   int errval;
   FILE *tmp;
   char lformat[80];
   double x[20];
   int    nc, p[20];
   char *c;
       
   lIndex = Index + ref;
   if ( lIndex->lines < 1 )
   {     printf(" No data to write = %d\n", lIndex->lines );
         return ref;
         }
   s = Index[ref].scan;
	strcpy(lformat, wformat);
   while ( (c=strstr (lformat, "\\n" )) != 0)
   {   *c++ = ' ';
       *c = '\n';
       }
   nc = 20;
   for (j=0; j<20; j++) p[j] = 0;
   nc = sscanf(wcolumns, "%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
                         p, p+1, p+2, p+3, p+4, p+5, p+6, p+7, p+8, p+9,
                         p+10, p+11, p+12, p+13, p+14, p+15, p+16, p+17, p+18, p+19);
   if ( nc < 0 ) 
	{   printf(" printing failed ; %d column(s) < 0, column(s)  \"%s\", format \"%s\"\n", nc, wcolumns, wformat);
	    return ref;
		 }
   tmp = outputfile(ref, s, name);
   printf(" printing %d column(s) \"%s\" with format \"%s\"\n", nc, wcolumns, wformat);
   fseek( inf, lIndex->cseek, SEEK_SET );
   do
   {    fgets( buffer, BUFSIZE, inf);
        if ( isdigit(buffer[0]) || buffer[0] == '-' || buffer[0] == '+' ) continue;
        if ( buffer[0] == '#' ) if (comments) fprintf( tmp,  "%s", buffer);
        } while ( ftell(inf) < lIndex->eseek );
   if (comments) fprintf( tmp, "#\n#spec2grace extracted columns \"%s\"\n", wcolumns);

   fseek( inf, lIndex->vseek, SEEK_SET );
   for ( i = 0; i < lIndex->lines; i++)
   {     fgets( buffer, BUFSIZE, inf);
         if ( buffer[0] == '#' ) 
         {  i--;
            if (comments) fprintf(tmp, "%s", buffer);
            if (buffer[1] == 'S' || buffer[1] == 'R') break; 
            continue;
            }
         if ( i==0 ) ncol = get_ncol( buffer, lIndex);
         for (j=0; j<nc; j++)
         {  x[j] = getdouble( buffer, p[j], &errval );
            if ( errval ) break;
            }
         if ( errval ) break;
         fprintf(tmp, lformat, x[0], x[1],x[2],x[3],x[4],x[5],x[6],x[7],x[8],x[9],
              x[10],x[11],x[12],x[13],x[14],x[15],x[16],x[17],x[18],x[19]);
         };
    if (tmp!=stdout) fclose(tmp);     
    return ref;    
    }

/* ---------------------------------------------------------------------- */
int mesh_edf( int ref, char *name  )
{  scanRef * lIndex;
   int i, j, s ;
   int errval;
   FILE *tmp;
   char *c;
   int sz, dx,dy;
   double *dd,dmi, dma;
   union { short int sh; char  c[2];} byteorder;
       
   lIndex = Index + ref;
   if ( lIndex->lines < 1 )
   {     printf(" No data to write = %d\n", lIndex->lines );
         return ref;
         }
   s = Index[ref].scan;
   xcol_used = xcol_select(ref);
   if (in_mesh==0)
   {   printf ("not hklmesh, can't export as image\n");
       return ref;
       }
   {  char file_name[50];
	   c = name;
      while ( *c == ' ') c++;
      if ( *c == '-' ) 
	   {   tmp = my_stdout; 
		    printf( " writing data for (i=%d, s=%d) on 'stdout'\n", ref+1, s); 
	       } 
	   else
	   {  if ( strlen (c) )
	      {   char *d = c + ( strlen (c) -1);
             while (*d == ' ' ||  *d == '\n' ||  *d == '\r') { *d=0; d--; }
			    }
		   if ( *c == '*' || *c == 0 ) 
	      {   sprintf(file_name,"scan%03d_%03d.edf", s, ref+1);
	          c = file_name;
			    }
	      tmp = fopen( c, "w" );
	      } 
	  }		 
   dx=++in_mesh;
   dy=(lIndex->lines+1)/dx;
   sz=dx*dy;
   dd=mesh_array(&dmi,&dma, lIndex); 
   if (dd==NULL) return ref;
   printf(" exporting mesh to  '%s' image : %d=%dx%d, float from %lf to %lf\n", 
     c,sz,dx,dy,dmi,dma);
   if (sz!=lIndex->lines) printf(" incomplete mesh (lines=%d) padded with 0\n", lIndex->lines);
   printf(" exporting mesh to  '%s' : %d = %d x %d float image\n", c,sz,dx,dy);
   fprintf(tmp,"{\nHeaderID = EH:000001:000000:000000\n\n");
   byteorder.sh=0x7FFF;
   if (byteorder.c[0]>0)  
   {   fprintf(tmp,"ByteOrder = HighByteFirst\n");
       }
   else      
   {   fprintf(tmp,"ByteOrder = LowByteFirst\n");
       }
   fprintf(tmp,"DataType = FloatValue\n");
   fprintf(tmp,"Dim_1  = %d\nDim_2  = %d\n", dx, dy);
   fflush(tmp);
   for (j=ftell(tmp); j<1000; j+=16)  fprintf(tmp,"%15s\n"," ");
   fflush(tmp);
   for (j=ftell(tmp); j<1021; j++)  fprintf(tmp," ");
   fprintf(tmp,"\n}\n");
   fflush(tmp);
   fseek(tmp,1024,0);
   for ( i=0 ; i < lIndex->lines; i++)
   {  float f=dd[i];
      fwrite(&f,sizeof(float),1,tmp);
      };
   for ( i=lIndex->lines; i<sz; i++)
   {  float v=0;
      fwrite(&v,sizeof(float),1,tmp);
      };
   fflush(tmp);
   dy=ftell(tmp);
   printf(" edf image  with %d header bytes and file size = %d\n",  1024 ,dy);
   if (tmp!=stdout) 
   {   fclose(tmp);
       }     
    free (dd);
    return ref;    
    }
/* ---------------------------------------------------------------------- */
int mesh_tiff( int ref, char *name  )
{  scanRef * lIndex;
   int i, j, s ;
   int errval;
   FILE *tmp;
   char *c;
   int sz, dx, dy;
   double *dd,dmi, dma;
   double dshift, dscale;
   unsigned int imin,imax;
   union { short int sh; char  c[2];} byteorder;
   struct 
   {  union { char c[2]; short unsigned u;} tag;
      union { char c[2]; short unsigned u;} typ;
      union { char c[4]; int l;}  cnt;
      union { char c[4]; short unsigned u[2]; int l; } val;
      } tif_ifd;
       
   lIndex = Index + ref;
   if ( lIndex->lines < 1 )
   {     printf(" No data to write = %d\n", lIndex->lines );
         return ref;
         }
   s = Index[ref].scan;
   xcol_used = xcol_select(ref);
   if (in_mesh==0)
   {   printf ("not hklmesh, can't export as image\n");
       return ref;
       }
   {  char file_name[50];
	   c = name;
      while ( *c == ' ') c++;
      if ( *c == '-' ) 
	   {   tmp = my_stdout; 
		    printf( " writing data for (i=%d, s=%d) on 'stdout'\n", ref+1, s); 
	       } 
	   else
	   {  if ( strlen (c) )
	      {   char *d = c + ( strlen (c) -1);
             while (*d == ' ' ||  *d == '\n' ||  *d == '\r') { *d=0; d--; }
			    }
		   if ( *c == '*' || *c == 0 ) 
	      {   sprintf(file_name,"scan%03d_%03d.tiff", s, ref+1);
	          c = file_name;
			    }
	      tmp = fopen( c, "w" );
	      } 
	  }		 
   dx=++in_mesh;
   dy=(lIndex->lines+1)/dx;
   sz=dx*dy;
   dd=mesh_array(&dmi,&dma, lIndex); 
   if (dd==NULL) return ref;
   if ((dma-dmi) > 0x0000FFFFL ) dshift = dmi; else dshift = 0;
   if (dma-dshift > 0x0000FFFFL ) dscale = 0x0000FFFFL / (dma-dshift); else dscale = 1;
   imin=dscale*(dmi-dshift);
   imax=dscale*(dma-dshift);
   printf(" exporting mesh to  '%s' image : %d=%dx%d, 16 bits grey from %d to %d\n", 
     c,sz,dx,dy,imin,imax);
   if (sz!=lIndex->lines) printf(" incomplete mesh (lines=%d) padded with 0\n", lIndex->lines);
   byteorder.sh=0x7FFF;
   if (byteorder.c[0]>0)  
   {   tif_ifd.tag.u=0x5555; 
       tif_ifd.typ.u=0x2a; 
       fwrite(&tif_ifd,4,1,tmp);  
       }
   else      
   {   tif_ifd.tag.u=0x4949; 
       tif_ifd.typ.u=0x2a; 
       fwrite(&tif_ifd,4,1,tmp);  
       }
   tif_ifd.cnt.l=8; 
   fwrite(&tif_ifd.cnt,4,1,tmp);  
   tif_ifd.typ.u=0x0a;                    /* tag counts */
   fwrite(&tif_ifd.typ,2,1,tmp);             
   tif_ifd.tag.u=0x100; tif_ifd.typ.u=3; tif_ifd.cnt.l=1; 
   tif_ifd.val.u[0] = dx;
   fwrite(&tif_ifd,sizeof(tif_ifd),1,tmp);  /* ImageWidth */
   tif_ifd.tag.u=0x101; tif_ifd.typ.u=3; tif_ifd.cnt.l=1; 
   tif_ifd.val.u[0] = dy;
   fwrite(&tif_ifd,sizeof(tif_ifd),1,tmp);  /* ImageLength */
   tif_ifd.tag.u=0x102; tif_ifd.typ.u=3; tif_ifd.cnt.l=1; 
   tif_ifd.val.u[0] = 16;
   fwrite(&tif_ifd,sizeof(tif_ifd),1,tmp);  /* BitsPerSample */
   tif_ifd.tag.u=0x103; tif_ifd.typ.u=3; tif_ifd.cnt.l=1; 
   tif_ifd.val.u[0] = 1;
   fwrite(&tif_ifd,sizeof(tif_ifd),1,tmp);  /* Compression */
   tif_ifd.tag.u=0x106; tif_ifd.typ.u=3; tif_ifd.cnt.l=1; 
   tif_ifd.val.u[0] = 0;
   fwrite(&tif_ifd,sizeof(tif_ifd),1,tmp);  /* PhotometricInterpretation */
   tif_ifd.tag.u=0x111; tif_ifd.typ.u=3; tif_ifd.cnt.l=1; 
   tif_ifd.val.u[0] = 0x0A + 10 * sizeof(tif_ifd) + 4;
   fwrite(&tif_ifd,sizeof(tif_ifd),1,tmp);  /* StripOffsets */
   tif_ifd.tag.u=0x116; tif_ifd.typ.u=4; tif_ifd.cnt.l=1; 
   tif_ifd.val.l = dy;
   fwrite(&tif_ifd,sizeof(tif_ifd),1,tmp);  /* RowPerStrip */
   tif_ifd.tag.u=0x117; tif_ifd.typ.u=4; tif_ifd.cnt.l=1; 
   tif_ifd.val.l = 2*sz;
   fwrite(&tif_ifd,sizeof(tif_ifd),1,tmp);  /* StripByteCounts */
   tif_ifd.tag.u=0x118; tif_ifd.typ.u=3; tif_ifd.cnt.l=1; 
   tif_ifd.val.u[0] = imin;
   fwrite(&tif_ifd,sizeof(tif_ifd),1,tmp);  /* MinSampleValue */
   tif_ifd.tag.u=0x119; tif_ifd.typ.u=3; tif_ifd.cnt.l=1; 
   tif_ifd.val.u[0] = imax;
   fwrite(&tif_ifd,sizeof(tif_ifd),1,tmp);  /* MaxSampleValue */
   tif_ifd.tag.u=0; tif_ifd.typ.u=0; tif_ifd.cnt.l=0; 
   tif_ifd.cnt.l = 0;
   fwrite(&tif_ifd,4,1,tmp);                /* zero to fill */
   fflush(tmp);
   dx=ftell(tmp);
   for ( i=0 ; i < lIndex->lines; i++)
   {  unsigned short v=dscale*(dd[i]-dshift);
      fwrite(&v,sizeof(unsigned short),1,tmp);
      };
   for ( i=lIndex->lines; i<sz; i++)
   {  unsigned short v=0;
      fwrite(&v,sizeof(unsigned short),1,tmp);
      };
   fflush(tmp);
   dy=ftell(tmp);
   printf(" tiff image shifted by %.2lf rescaled %.2lf with %d header bytes and file size = %d\n", 
    dshift  ,dscale, dx,dy);
   if (tmp!=stdout) 
   {   fclose(tmp);
       }     
    free (dd);
    return ref;    
    }
/* ---------------------------------------------------------------------- */
static double *mesh_array(double *dmi,double *dma, scanRef * lIndex)
{  int i, errval;
   double *dd;
   fseek( inf, lIndex->vseek, SEEK_SET );
   dd=(double *) malloc(sizeof(double)*lIndex->lines);
   memset(dd,0,sizeof(double)*lIndex->lines);
   if (dd==NULL)
   {  printf("can't allocated array\n");
      return NULL;
      }
   for ( i = 0; i < lIndex->lines; i++)
   {   double y,z;
       fgets( buffer, BUFSIZE, inf);
         if ( buffer[0] == '#' ) 
         {   i--;
             if (buffer[1] == 'S' || buffer[1] == 'R') break; 
             continue;
             }
         if ( i==0 ) ncol = get_ncol( buffer, lIndex);
         y = getdouble( buffer, ycol, &errval ) + ZERO ;
         if ( normalize ) 
         {   z = getdouble( buffer, zcol, &errval );
             if ( z >= 1 ) 
             {   y *= ref_norm/z;
                 }
             }
         if(y>*dma || i==0) *dma=y;
         if(y<*dmi || i==0) *dmi=y;
         dd[i]=y;
         };
    return dd;
    }
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/*     the following procedure interact with xmgrace                      */
/* ---------------------------------------------------------------------- */
#ifndef NOGRACE
void grace_atexit( void);
int graceGraph = 0;
int GRACEPRINTF(char *f, ...)
{  char grace_dbg[256];
    va_list argptr;
    va_start (argptr, f);
	 vsprintf (grace_dbg,f, argptr);
	 va_end (argptr);
    GracePrintf(grace_dbg);
#ifdef DEBUG
    fprintf(stderr,grace_dbg);
#endif
    }  

int graceOrGnuplotSendScan( int ref )
{
   /* recherche si MESH */
   scanRef *lIndex;
   lIndex = Index + ref;
   fseek( inf, lIndex->cseek, SEEK_SET );
   mesh=0;
   do
   {    fgets( buffer, BUFSIZE, inf);
        if ( (buffer[0] == '#') && (buffer[1] == 'S'))
            if (strstr(buffer, "mesh") != NULL) {mesh=1; break; }
   } while ( ftell(inf) < lIndex->eseek );
   if (mesh ==0) gracesend_scan(ref);
   else
      {
      char hklmesh[256], H, K, L;
      float f1,f2,f3;
      sscanf(strstr(buffer, "mesh"), "%s %c %g %g %g %c", hklmesh, &H, &f1, &f2, &f3, &K);
      x1col_select= 1;
      if (H == 'K') { x1col_select= 2; x2col_select= 3; }
      else { x2col_select=2; if (K == 'L')  x2col_select=3; }
      gnuplotsend_scan(ref);
      }
}

#define GRACEBUFFERSIZE  2048  
int gracesend_scan( int ref )
{
   scanRef * lIndex;
   int i;
   int errval;
   FILE *tmp;
   FILE *file;
   double x, y, z, e;
   char fname[50];
   char legend[256];
   
    /* Start Grace with a buffer size and open the pipe */
    if ( !GraceIsOpen() )
    {  GraceRegisterErrorFunction (graceerror_function);
	    if (GraceOpen( GRACEBUFFERSIZE  ) == -1) 
       {   fprintf(stderr, "Can't run Grace.\n");
           return ref;
           }
       gracewin_size();
       atexit( grace_atexit);
       } 
       
   lIndex = Index + ref;
   fseek( inf, lIndex->vseek, SEEK_SET );
   if ( lIndex->lines <= 1 )
   {     printf(" No data to plot lines = %d\n", lIndex->lines );
         return ref;
         }
   strcpy(fname, prgname);
   strcat(fname,".dat");
   tmp = fopen( fname, "w" );
   if ( !tmp) {
      fprintf(stderr,"unable to write file '%s'\n", fname);
      return ref;
      }
   
   xcol_used = xcol_select(ref);
   if ( gracewinlegend[0]==0)
   {  sprintf(legend,"%s #S%d", inf_name, lIndex->scan);
      }
   else if (gracewinlegend[0]=='+')
   {  sprintf(legend,"%s #S%d ", inf_name, lIndex->scan);
      strcat(legend,spec_tag(ref,gracewinlegend[1]));
      }
   else
   {  sprintf(legend,"%s ", inf_name);
      strcat(legend,spec_tag(ref,gracewinlegend[0]));
      }
   for ( i = 0; i < lIndex->lines; i++)
   {     fgets( buffer, BUFSIZE, inf);
         if ( buffer[0] == '#' )
         {   i--;
             continue;
             }
         if ( i==0 ) ncol = get_ncol( buffer, lIndex);
         x = getdouble( buffer, xcol_used, &errval );
         if ( errval ) break;
         y = getdouble( buffer, ycol, &errval ) + ZERO ;
         if ( errorbar ) 
         {   if ( y >= 1) e = sqrt(y); else e = 0;
             }
         if ( normalize ) 
         {   z = getdouble( buffer, zcol, &errval );
             if ( z >= 1 ) 
             {   y *= ref_norm / z;
                 if (errorbar) e = sqrt( e*e + y*y/z)/z;
                 }
             }
         if ( errorbar ) fprintf(tmp, "%lg %lg %lg\n", x, y, e);
         else fprintf(tmp, "%lg %lg\n", x, y);
         };
    fclose(tmp);     
    if (autopurge) GRACEPRINTF("Kill G%d\nG%d on\nWith G%d", graceGraph, graceGraph, graceGraph);  
    file = fopen( "spec2grace.par", "r" );
    if ( file != NULL )
    {    printf("Automatic xmgr parameter file : spec2grace.par\n");
         GRACEPRINTF("GETP \"spec2grace.par\"\n");
         fclose(file);
         }
    if (autopurge || autoscale)  GRACEPRINTF( "AUTOSCALE ONREAD XYAXES\n");
    else GRACEPRINTF( "AUTOSCALE ONREAD NONE\n");
    if (errorbar ) GRACEPRINTF( "READ XYDY \"%s\"\n", fname );
    else           GRACEPRINTF( "READ XY \"%s\"\n", fname );
    GracePrintf("S_ COMMENT \"%s #S%d\"\n", inf_name, lIndex->scan);
    GracePrintf("S_ LEGEND \"%s\"\n", legend);
    GracePrintf( "G%d TYPE %s\n", graceGraph, ( logscale == 0 ? "XY" : "LOGY" ) );
    file = fopen( "spec2grace.bat", "r" );
    if ( file != NULL )
    {    printf("Automatic xmgr parameters file : spec2grace.bat\n");
         GracePrintf("GETP \"spec2grace.bat\"\n");
         fclose(file);
         }
    if (titleauto) GracePrintf( "SUBTITLE \"%s #S%d\"\n", inf_name, lIndex->scan);
    GracePrintf( "LEGEND %s\n", legendauto?"ON":"OFF");
	 GRACEPRINTF( "redraw\n");      
	 GRACEPRINTF( "UPDATEALL\n");   
    GraceFlush();    
#ifdef DEBUG
    fprintf(stderr,"GraceFlush\n");
#endif    
    return ref;    
    }


int gnuplotsend_scan( int ref )
{
   scanRef *lIndex;
   int i;
   int errval;
   FILE *tmp;
   FILE *file;
   double x1, x1Old, x2, x2Old, y, z, e;
   char fname[50];
   char legend[256];
   char cmd[256];
          
   lIndex = Index + ref;
   fseek( inf, lIndex->vseek, SEEK_SET );
   if ( lIndex->lines <= 1 )
   {     printf(" No data to plot lines = %d\n", lIndex->lines );
         return ref;
         }
   strcpy(fname, prgname);
   strcat(fname,".dat");
   tmp = fopen( fname, "w" );
   if ( !tmp) {
      fprintf(stderr,"unable to write file '%s'\n", fname);
      return ref;
      }
   
   x1col_used = x1col_select;
   x2col_used = x2col_select;
   if ( gracewinlegend[0]==0)
   {  sprintf(legend,"%s #S%d", inf_name, lIndex->scan);
      }
   else if (gracewinlegend[0]=='+')
   {  sprintf(legend,"%s #S%d ", inf_name, lIndex->scan);
      strcat(legend,spec_tag(ref,gracewinlegend[1]));
      }
   else
   {  sprintf(legend,"%s ", inf_name);
      strcat(legend,spec_tag(ref,gracewinlegend[0]));
      }
   for ( i = 0; i < lIndex->lines; i++)
   {     fgets( buffer, BUFSIZE, inf);
         if ( buffer[0] == '#' )
         {   i--;
             continue;
             }
         if ( i==0 ) ncol = get_ncol( buffer, lIndex);
         x1 = getdouble( buffer, x1col_used, &errval );
         if ( errval ) break;
         x2 = getdouble( buffer, x2col_used, &errval );
         if ( errval ) break;
         y = getdouble( buffer, ycol, &errval ) + ZERO ;
         if ( errorbar ) 
         {   if ( y >= 1) e = sqrt(y); else e = 0;
             }
         if ( normalize ) 
         {   z = getdouble( buffer, zcol, &errval );
             if ( z >= 1 ) 
             {   y *= ref_norm / z;
                 if (errorbar) e = ref_norm*sqrt( e*e + ref_norm*y*y/z)/z;
                 }
             }
         if ( logscale !=0 ) y = log(y);
         if (x2 != x2Old) fprintf(tmp, "\n");
         x2Old=x2;
         if ( errorbar ) fprintf(tmp, "%lg %lg %lg %lg\n", x1, x2, y, e);
         else fprintf(tmp, "%lg %lg %lg\n", x1, x2, y);
         };
    fclose(tmp);     
    printf("MESH\n");
    sprintf(cmd, "gnuplot %s", GNUPLOT_CMD_FILE);
    system(cmd);
    return ref;    
    }
/* ------------------------------------------------------ */
int gracesend_print( char * name )
{  
   if ( mesh == 1) { system("lp spec2graceGnuPlotOutput.ps"); return(0); }
   if ( graceprint[0] == 0)
         GracePrintf( "%s\n", GRACEPRINT);
	else  GracePrintf( "%s\n", graceprint);
	GracePrintf( "PRINT\n");
   }
/* ------------------------------------------------------ */
int gracesend_batch( char * name )
{    FILE *file;
     file = fopen( name, "r" );
     if ( file == NULL )
     {     printf(" xmgr batch file %s not found\n", name);
           return -1;
           }
     printf("Using xmgr batch file %s\n", name);
          GracePrintf("READ BATCH \"%s\"\n", name);
     fclose(file);
     return 0;
     }
/* ---------------------------------------------------------------------- */
int gracesend_parameter( char * name )
{    FILE *file;
     file = fopen( name, "r" );
     if ( file == NULL )
     {     printf(" xmgr parameter file %s not found\n", name);
           return -1;
           }
     printf("Using xmgr parameter file %s\n", name);
     GracePrintf("GETP \"%s\"\n", name);
     fclose(file);
     return 0;
     }
/* ---------------------------------------------------------------------- */
void gracewin_size(void)
{  if ( GraceIsOpen() && graceWin[0] )
        GracePrintf( "PAGE RESIZE %s\nREDRAW\n", graceWin);
    }
/* ---------------------------------------------------------------------- */
void grace_atexit(void)
{   gracewin_size();    
    /* to ensure that the X server will not receive the exit command
       if xmgrace has been killed but the pipe remains still open
       bug???  corrected in new X server */
    if ( GraceIsOpen() )  GraceClose();
    }    
/* ---------------------------------------------------------------------- */
void graceerror_function(const char *msg)
{   fprintf(stderr, "grace library message : \"%s\"\n", msg);
    }
/* ---------------------------------------------------------------------- */
#endif

